/*
 * Workspaces.cpp
 * Open BeOS version alpha 1 by Francois Revol revol@free.fr
 *
 * Workspaces window trick found by Michael "Minox" Paine.
 * (using B_ALL_WORKSPACES as flags in BWindow)
 * Found out that using 0xffffffff as Flags was causing the window not to close on Alt-W
 * hey Workspaces get Flags of Window 0
 * gives 0x00008080 which makes it.
 *
 * Modified by matt madia 12/29/03,  
 *	now allows for command line arguments:
 *	"--notitle" just removes title bar.  keeps border and ability to resize
 *	"--noborder" removes title, border, and ability to resize
 *	"--acceptfirstclick" forces workspace switch on first click
 *	"--avoidfocus" Prevents the window from being the target of keyboard events.
 *	all can be set simutaneously or in any pair 
 *	(even though --notitle is not necessary when --noborder is used) 
 *
 * oliver "Madison" kohl + matt madia 01/05/04,
 *	maintains window position correctly!
 *  	automatic feature, need not be enabled.
 *
 *
 */

/*
	possible future additions.
	"--sendbehind"   sendBehind() after switching workspaces.
*/

#include <Application.h>
#include <Alert.h>
#include <File.h>
#include <Window.h>
#include <Screen.h>
#include <Entry.h>
#include <Roster.h>
#include <Path.h>
#include <FindDirectory.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

class WorkspacesPreferences {
	public:
		WorkspacesPreferences();
		virtual ~WorkspacesPreferences();

		BRect WindowFrame() const { return fWindowFrame; }
		BRect ScreenFrame() const { return fScreenFrame; }
		BRect PreviousWindowFrame() const { return previousWindowFrame; }
		BRect PreviousScreenFrame() const { return previousScreenFrame; }
		bool LoadedDifferentResolution() const { return loadedDifferentResolution; }

		void UpdateScreenFrame();
		void SetWindowFrame(BRect);
		void SetPreviousScreenFrame(BRect);
		void SetPreviousWindowFrame(BRect);
		void SetLoadedDifferentResolution(bool);

	private:
		static const char kWorkspacesSettingFile[];
		static const char kWorkspacesResolutionFile[];
		BRect fWindowFrame, fScreenFrame;
		BRect previousWindowFrame, previousScreenFrame;
		bool loadedDifferentResolution;
};

class WorkspacesWindow : public BWindow {
	public:
		WorkspacesWindow(WorkspacesPreferences *fPreferences);
		virtual ~WorkspacesWindow();
	
		virtual void ScreenChanged(BRect frame,color_space mode);
		virtual void FrameMoved(BPoint origin);
		virtual void FrameResized(float width, float height);
		virtual void Zoom(BPoint origin, float width, float height);
		virtual void MessageReceived(BMessage *msg);
		virtual bool QuitRequested(void);
		
	private:
		WorkspacesPreferences *fPreferences;
};

class WorkspacesApp : public BApplication {
	public:
		WorkspacesApp();
		virtual ~WorkspacesApp();

		virtual void AboutRequested(void);
		virtual void ArgvReceived(int32 argc, char **argv);
		virtual void ReadyToRun(void);
		virtual bool WorkspacesApp::QuitRequested(void);
		virtual void Usage(void);
		virtual void Version(void);
		virtual void SuggestHelpMessage(void);

	private:
		static const char kWorkspacesAppSig[];
		WorkspacesWindow *fWorkspacesWindow;
		char *prog_name;
};

const char WorkspacesApp::kWorkspacesAppSig[] = "application/x-vnd.Be-WORK";
const char WorkspacesPreferences::kWorkspacesSettingFile[] = "Workspace_data";
const char WorkspacesPreferences::kWorkspacesResolutionFile[] = "Workspace_resolution";

WorkspacesPreferences::WorkspacesPreferences()
{
	UpdateScreenFrame();
	BPath workspaceResolutionPath;  
	BScreen screen;

	if (find_directory(B_USER_SETTINGS_DIRECTORY,
		&workspaceResolutionPath) == B_OK) {
		workspaceResolutionPath.Append(kWorkspacesResolutionFile);
		BFile workspaceResolutionFile(workspaceResolutionPath.Path(),
			 B_READ_ONLY);
		if (workspaceResolutionFile.InitCheck() == B_OK 
			&& workspaceResolutionFile.Read(&fScreenFrame, 
					sizeof(BRect)) == sizeof(BRect)) {
			BPath workspaceDataPath; 
			if (find_directory(B_USER_SETTINGS_DIRECTORY,
				&workspaceDataPath) == B_OK) {
				workspaceDataPath.Append(kWorkspacesSettingFile);
				BFile workspaceDataFile(workspaceDataPath.Path(), B_READ_ONLY);
				if (workspaceDataFile.InitCheck() == B_OK 
					&& workspaceDataFile.Read(&fWindowFrame,
							 sizeof(BRect)) == sizeof(BRect)) {
					if ( fScreenFrame != screen.Frame() ){
						loadedDifferentResolution = true;
						return;
					}
					if (screen.Frame().right >= fWindowFrame.right
  						  && screen.Frame().bottom >= fWindowFrame.bottom){
						loadedDifferentResolution = false;
						return;
					}
				}
			}
			if ( fScreenFrame != screen.Frame() ){
				loadedDifferentResolution = true;
				return;
			}
		}
	}
	fWindowFrame = fScreenFrame;
	fWindowFrame.OffsetBy(-10, -10);
	fWindowFrame.left = fWindowFrame.right - 160;
	fWindowFrame.top = fWindowFrame.bottom - 120;
}


WorkspacesPreferences::~WorkspacesPreferences()
{
	// write Workspace_data  -- window size
	BPath workspaceDataPath;
	if (find_directory(B_USER_SETTINGS_DIRECTORY,&workspaceDataPath) < B_OK)
		return;

	workspaceDataPath.Append(kWorkspacesSettingFile);

	BFile workspaceDataFile(workspaceDataPath.Path(),
		 B_WRITE_ONLY | B_CREATE_FILE);
	if (workspaceDataFile.InitCheck() == B_OK)
		workspaceDataFile.Write(&fWindowFrame, sizeof(BRect));
	
	// write Workspace_resolution -- last resolution
	BPath workspaceResolutionPath;
	if (find_directory(B_USER_SETTINGS_DIRECTORY,
		&workspaceResolutionPath) < B_OK)
		return;

	workspaceResolutionPath.Append(kWorkspacesResolutionFile);

	BFile workspaceResolutionFile(workspaceResolutionPath.Path(),
		 B_WRITE_ONLY | B_CREATE_FILE);
	if (workspaceResolutionFile.InitCheck() == B_OK)
		workspaceResolutionFile.Write(&fScreenFrame, sizeof(BRect));
}

void WorkspacesPreferences::SetLoadedDifferentResolution(bool b) 
{
	loadedDifferentResolution = b;
}

void
WorkspacesPreferences::SetPreviousScreenFrame(BRect f)
{
	previousScreenFrame = f;
}

void
WorkspacesPreferences::SetPreviousWindowFrame(BRect f)
{
	previousWindowFrame = f;
}


void
WorkspacesPreferences::UpdateScreenFrame()
{
	BScreen screen;
	fScreenFrame = screen.Frame();
}

void
WorkspacesPreferences::SetWindowFrame(BRect f)
{
	fWindowFrame = f;

	UpdateScreenFrame();
}

//	#pragma mark -


// here is the trick :)
#define B_WORKSPACES_WINDOW 0x00008000
WorkspacesWindow::WorkspacesWindow(WorkspacesPreferences *preferences)
	: BWindow(preferences->WindowFrame(), "Workspaces", B_TITLED_WINDOW_LOOK,
 			B_NORMAL_WINDOW_FEEL, B_WORKSPACES_WINDOW | B_AVOID_FRONT, B_ALL_WORKSPACES)
{
	fPreferences = preferences;
	// record initial resolution and window position as old,
	// for first call to ScreenChanged()
	fPreferences->SetPreviousScreenFrame(fPreferences->ScreenFrame()); 
	fPreferences->SetPreviousWindowFrame(Frame());  
	if ( fPreferences->LoadedDifferentResolution() ) {  
		BScreen screen;
		fPreferences->SetPreviousWindowFrame(fPreferences->WindowFrame());
		ScreenChanged(screen.Frame(), screen.ColorSpace());
	}	
}

WorkspacesWindow::~WorkspacesWindow()
{
	delete fPreferences;
}


void
WorkspacesWindow::ScreenChanged(BRect rect, color_space mode)
{
	//`rect`  		new screen dimensions
	//`Frame()`		new ws window location

	// need to be stored locally, since PreviousWindowFrame() will get modified.
	// well, it would if FrameMoved() weren't called before ScreenChanged().
	BRect oldScreenFrame = fPreferences->PreviousScreenFrame();
	BRect oldWindowFrame = fPreferences->PreviousWindowFrame();

	// adjust horizontal position
	if (oldWindowFrame.right > oldScreenFrame.right/2)  
		MoveTo(rect.right - (oldScreenFrame.right - oldWindowFrame.left),
			oldWindowFrame.top); 

	// adjust vertical position
	if (oldWindowFrame.bottom > oldScreenFrame.bottom/2) 
		MoveTo(Frame().left, 
			rect.bottom - (oldScreenFrame.bottom - oldWindowFrame.top));

	// call these at the end of the function, so it's ready for next call.
	// FLAW: moving the workspace window after a ScreenChanged will use these values.
	fPreferences->SetPreviousScreenFrame(rect);  
	fPreferences->SetPreviousWindowFrame(Frame());

}
void
WorkspacesWindow::FrameMoved(BPoint origin)
{
	fPreferences->SetWindowFrame(Frame());
	// When switching workspaces, FrameMoved is called before ScreenChanged.
	//  prevents ScreenChange correctly repositioning after a 
	//  workspaces window repositition.
	// Must find a way to SetPreviousWindowFrame(Frame()) in FrameMoved(),
	//  but not when FrameMoved is called upon switching workspaces.
	//fPreferences->SetPreviousWindowFrame(Frame());
}


void
WorkspacesWindow::FrameResized(float width, float height)
{
	fPreferences->SetWindowFrame(Frame());
	fPreferences->SetPreviousWindowFrame(Frame());
}


void 
WorkspacesWindow::Zoom(BPoint origin, float width, float height)
{
	BScreen screen;
	origin = screen.Frame().RightBottom();
	origin.x -= 10 + fPreferences->WindowFrame().Width();
	origin.y -= 10 + fPreferences->WindowFrame().Height();

	MoveTo(origin);
}


void
WorkspacesWindow::MessageReceived(BMessage *msg)
{
	if (msg->what == 'DATA') { // Drop from Tracker
		entry_ref ref;
		for (int i = 0; (msg->FindRef("refs", i, &ref) == B_OK); i++)
			be_roster->Launch(&ref);
	} else
		BWindow::MessageReceived(msg);
}


bool
WorkspacesWindow::QuitRequested(void)
{
	be_app->PostMessage(B_QUIT_REQUESTED);
	return true;
}


//	#pragma mark -


WorkspacesApp::WorkspacesApp() : BApplication(kWorkspacesAppSig)
{	
	fWorkspacesWindow = new WorkspacesWindow(new WorkspacesPreferences());
}


WorkspacesApp::~WorkspacesApp()
{
}


void
WorkspacesApp::AboutRequested(void)
{
	// blocking !! 
	// the original does the same by the way =)
	(new BAlert("about", "OpenBeOS Workspaces\n\tby François Revol,
		 Axel Dörfler, Matt Madia.\n\noriginal Be version by Robert Polic", "Big Deal"))->Go();
}

void
WorkspacesApp::Usage(void)
{
	(void)fprintf(stdout, 
"%s OBOS (http://www.openbeos.org/)

Usage: %s [OPTION]... [USERNAME]

  --notitle   	  		title bar removed.  border and resize kept.
  --noborder   	  		title, border, and resize removed.
  --acceptfirstclick  	forces workspace switch on first click.
  --avoidfocus   		prevents the window from being the target of keyboard events.
  --help      			display this help and exit
  --version   			output version information and exit

Print information for USERNAME, or the current user.
	
", prog_name, prog_name	);
	Quit();
}

void
WorkspacesApp::Version(void)
{
	(void)fprintf(stdout, "%s OBOS (http://www.openbeos.org/)\n", prog_name);
	Quit();
}

void
WorkspacesApp::SuggestHelpMessage(void)
{
	(void)fprintf(stdout, "Try `%s --help' for more information.\n", prog_name);
	Quit();
}

void
WorkspacesApp::ArgvReceived(int32 argc, char **argv)
{
	prog_name = argv[0];
	// we are told to switch to a specific workspace
	if (argc > 1 && argv[1][0] != '-') {
		// get it from the command line...
		activate_workspace(atoi(argv[1]));
		// if the app is running, don't quit
		// but if it isn't, cancel the complete run, so it doesn't
		// open any window
		if (IsLaunching())
			Quit();
	}
	// evaluate arguments
	for (int i = 1;  i < argc;  i++)
	{
		if (! strcmp(argv[i], "--notitle")) {
			fWorkspacesWindow->SetLook(B_MODAL_WINDOW_LOOK);
			fWorkspacesWindow->SetFlags(B_OUTLINE_RESIZE);
		} else if (! strcmp(argv[i], "--noborder")) 
			fWorkspacesWindow->SetLook(B_NO_BORDER_WINDOW_LOOK);
		else if (! strcmp(argv[i], "--acceptfirstclick"))
			fWorkspacesWindow->SetFlags(B_WILL_ACCEPT_FIRST_CLICK);
		else if (! strcmp(argv[i], "--avoidfocus"))
			fWorkspacesWindow->SetFlags(B_AVOID_FOCUS);
		else if (! strcmp(argv[i], "--version")) 
			Version();
		else if (! strcmp(argv[i], "--help"))
			Usage();
		else if ( argv[i][0] == '-')
			SuggestHelpMessage();
	}
}


void
WorkspacesApp::ReadyToRun(void)
{
	(fWorkspacesWindow)->Show();
}


bool
WorkspacesApp::QuitRequested(void)
{
	return true;
}


//	#pragma mark -


int
main(int32 argc, char **argv)
{			
	new WorkspacesApp();
	be_app->Run();
	delete be_app;
	return 0;
}
